"""
board.py - Minesweeper board implementation

March 2018, Lewis Gaul

Exports:
Board (class)
    Representation of a minesweeper board.
Minefield (class)
    Representation of the positioning of mines in minesweeper board.
"""

__all__ = ("Board", "Minefield")

import random as rnd
from typing import Any, Collection, Dict, Iterable, List, Optional, Union

from ..shared import utils
from ..types import CellContentsType, CellFlag, CellMineType, CellNum, CellUnclicked
from ..typing import Coord_T


class Board(utils.Grid):
    """
    Representation of a minesweeper board. To be filled with instances of
    CellContentsType.
    """

    def __init__(self, x_size: int, y_size: int):
        """
        Arguments:
        x_size (int > 0)
            The number of columns.
        y_size (int > 0)
            The number of rows.
        """
        super().__init__(x_size, y_size, fill=CellUnclicked())

    def __repr__(self):
        return f"<{self.x_size}x{self.y_size} board>"

    def __str__(self):
        return super().__str__(mapping={CellNum(0): "."})

    def __setitem__(self, key, value):
        if not isinstance(value, CellContentsType):
            raise TypeError("Board can only contain CellContentsType instances")
        else:
            super().__setitem__(key, value)

    @classmethod
    def from_2d_array(cls, array):
        """
        Create a minesweeper board from a 2-dimensional array of string
        representations for cell contents.

        Arguments:
        array ([[str|int, ...], ...])
            The array to create the board from.

        Return:
            The created board.

        Raises:
        ValueError
            - Invalid string representation of cell contents.
        """
        grid = utils.Grid.from_2d_array(array)
        board = cls(grid.x_size, grid.y_size)
        for c in grid.all_coords:
            if type(grid[c]) is int:
                board[c] = CellNum(grid[c])
            elif type(grid[c]) is str and len(grid[c]) == 2:
                char, num = grid[c]
                board[c] = CellMineType.get_class_from_char(char)(int(num))
            elif grid[c] != CellUnclicked.char:
                raise ValueError(
                    f"Unknown cell contents representation in cell {c}: " f"{grid[c]}"
                )
        return board

    def reset(self):
        """Reset the board to the initial state."""
        for c in self.all_coords:
            self[c] = CellUnclicked()


class Minefield(utils.Grid):
    """
    Grid representation of a minesweeper minefield, with each cell containing
    an integer representing the number of mines in that cell.
    """

    def __init__(
        self,
        x_size: int,
        y_size: int,
        *,
        mines: Union[int, Iterable[Coord_T]],
        per_cell: int = 1,
        safe_coords: Optional[Iterable[Coord_T]] = None,
    ):
        """
        :param x_size:
            Number of columns in the grid.
        :param y_size:
            Number of rows in the grid.
        :param mines:
            Either the number of mines to randomly place, or an iterable of
            coordinates to place mines at.
        :param per_cell:
            Maximum number of mines per cell.
        :param safe_coords:
            Optionally specify coordinates that should not contain a mine when
            filling the minefield. Ignored if a list of mine coords is passed
            in.
        :raise ValueError:
            If the number of mines is too high to fit in the grid or if a list
            of mine coordinates is supplied and the number of mines in a cell
            exceeds the max per cell value.
        """
        super().__init__(x_size, y_size)
        # Maximum number of mines per cell.
        self.per_cell: int = per_cell
        # Number of mines.
        self.nr_mines: int
        # List of cell coordinates containing mines.
        self.mine_coords: List[Coord_T]
        # The successfully completed board for the minefield.
        self.completed_board: Board
        # Groups of cells that form the board openings.
        self.openings: Collection[Collection[Coord_T]]
        # The 3bv of the minefield.
        self.bbbv: int

        if isinstance(mines, int):
            self.nr_mines = mines
            mine_coords = self._choose_mine_coords(safe_coords)
        else:
            mine_coords = list(mines)
            self.nr_mines = len(mine_coords)

        for c in mine_coords:
            if self[c] == self.per_cell:
                raise ValueError(
                    f"Too many mines at coord {c} with max per cell of {self.per_cell}"
                )
            self[c] += 1
        self.mine_coords = mine_coords
        self.completed_board = self._calc_completed_board()
        self.openings = self._find_openings()
        self.bbbv = self._calc_3bv()

    def __repr__(self):
        mines_str = f" with {self.nr_mines} mines" if self.nr_mines else ""
        return f"<{self.x_size}x{self.y_size} minefield{mines_str}>"

    @classmethod
    def from_grid(cls, grid: utils.Grid, *, per_cell: int = 1) -> "Minefield":
        """
        Create a minefield with a grid showing where mines are to lie.

        :param grid:
            The grid of mines.
        :param per_cell:
            Maximum number of mines per cell. Should not be exceeded by any of
            the values in the grid.
        :return:
            The created minefield.
        """
        mine_coords = []
        for c in grid.all_coords:
            for _ in range(grid[c]):
                mine_coords.append(c)

        return cls(grid.x_size, grid.y_size, mines=mine_coords, per_cell=per_cell)

    @classmethod
    def from_2d_array(cls, array: List[List[int]], *, per_cell: int = 1) -> "Minefield":
        """
        See minegauler.core.utils.Grid and Minefield.from_grid().
        """
        return cls.from_grid(utils.Grid.from_2d_array(array), per_cell=per_cell)

    @classmethod
    def from_json(cls, dict_: Dict[str, Any]) -> "Minefield":
        """
        Create a minefield instance from a JSON encoding.

        :param dict_:
            The dictionary obtained from decoding JSON. Must contain the
            following fields: 'x_size', 'y_size', 'mine_coords'.
        :raise ValueError:
            If the dictionary is missing required fields.
        """
        try:
            return cls(
                dict_["x_size"],
                dict_["y_size"],
                mines=[tuple(c) for c in dict_["mine_coords"]],
                per_cell=dict_.get("per_cell", 1),
            )
        except KeyError as e:
            raise ValueError(
                "Missing key in dictionary when trying to create minefield"
            ) from e

    def to_json(self) -> Dict[str, Any]:
        """Convert to a JSON-serialisable format."""
        return dict(
            x_size=self.x_size,
            y_size=self.y_size,
            mine_coords=self.mine_coords,
            per_cell=self.per_cell,
        )

    def _choose_mine_coords(
        self, safe_coords: Optional[Iterable[Coord_T]] = None
    ) -> List[Coord_T]:
        """
        Randomly choose coordinates for mines to be in.

        :param safe_coords:
            Optionally specify coordinates that should not contain a mine when
            filling the minefield. Ignored if a list of mine coords is passed
            in.
        :return:
            A list of randomly chosen mine coords.
        :raise ValueError:
            If the number of mines is too high to fit in the grid.
        """
        safe_coords = set(safe_coords) if safe_coords else None
        self.check_enough_space(
            x_size=self.x_size,
            y_size=self.y_size,
            mines=self.nr_mines,
            per_cell=self.per_cell,
            nr_safe_cells=len(safe_coords) if safe_coords else 1,
        )

        # Get a list of coordinates which can have mines placed in them.
        if safe_coords is None:
            avble_coords = self.all_coords.copy()
        else:
            avble_coords = [c for c in self.all_coords if c not in safe_coords]
        # Make sure there is at least one safe cell.
        if len(avble_coords) == len(self.all_coords):
            avble_coords.pop(rnd.randint(0, len(avble_coords) - 1))
        avble_coords *= self.per_cell
        rnd.shuffle(avble_coords)
        return avble_coords[: self.nr_mines]

    @staticmethod
    def check_enough_space(
        *, x_size: int, y_size: int, mines: int, per_cell: int, nr_safe_cells: int = 1
    ) -> None:
        """
        Check there is enough space in the grid for the mines.

        :param x_size:
            Number of columns in the grid.
        :param y_size:
            Number of rows in the grid.
        :param mines:
            The number of mines.
        :param per_cell:
            Maximum number of mines per cell.
        :param nr_safe_cells:
            The number of cells to leave safe.
        :raise ValueError:
            If the number of mines is too high to fit in the grid.
        """
        mine_spaces = x_size * y_size - nr_safe_cells
        if mines > mine_spaces * per_cell:
            raise ValueError(
                f"Number of mines too high ({mines}) for grid with {mine_spaces} spaces "
                f"and only up to {per_cell} allowed per cell"
            )

    def cell_contains_mine(self, coord: Coord_T) -> bool:
        """
        Return whether a cell contains at least one mine.

        :param coord:
            The coordinate to check.
        :return:
            Whether the cell contains a mine.
        """
        return self[coord] > 0

    def _calc_completed_board(self) -> Board:
        """
        Create the completed board with the flags and numbers that should be
        seen upon game completion.
        """
        completed_board = Board(self.x_size, self.y_size)
        completed_board.fill(CellNum(0))
        for c in self.all_coords:
            mines = self[c]
            if mines > 0:
                completed_board[c] = CellFlag(mines)
                for nbr in self.get_nbrs(c):
                    # For neighbouring cells that don't contain mines, increment
                    #  their number.
                    if not self.cell_contains_mine(nbr):
                        completed_board[nbr] += mines
        return completed_board

    def _find_openings(self) -> List[List[Coord_T]]:
        """
        Find the openings of the board.

        A list of openings is stored, each represented as a list of
        coordinates belonging to that opening. Note that each cell
        cannot belong to multiple openings.
        """
        openings = []
        blanks_to_check = {
            c for c in self.all_coords if self.completed_board[c] == CellNum(0)
        }
        while blanks_to_check:
            orig_coord = blanks_to_check.pop()
            # If the coordinate is part of an opening and hasn't already been
            #  considered, start a new opening.
            opening = {orig_coord}  # Coords belonging to the opening
            check = {orig_coord}  # Coords whose neighbours need checking
            while check:
                coord = check.pop()
                nbrs = set(self.get_nbrs(coord))
                check |= {
                    c for c in nbrs - opening if self.completed_board[c] == CellNum(0)
                }
                opening |= nbrs
            openings.append(sorted(opening))
            blanks_to_check -= opening
        return openings

    def _calc_3bv(self) -> int:
        """Calculate the 3bv of the board."""
        clicks = len(self.openings)
        exposed = len({c for opening in self.openings for c in opening})
        clicks += self.x_size * self.y_size - len(set(self.mine_coords)) - exposed
        return clicks



"""
formatter.py - Format bot messages

February 2020, Lewis Gaul
"""

__all__ = (
    "format_highscores",
    "format_highscore_times",
    "format_kwargs",
    "format_matchups",
    "format_player_highscores",
)

import time
from typing import Iterable, List, Mapping, Optional, Tuple

from minegauler.shared import highscores as hs

from .utils import Matchup


def format_highscores(highscores: Iterable[hs.HighscoreStruct]) -> str:
    return format_highscore_times([(h.name, h.elapsed) for h in highscores])


def format_highscore_times(highscores: Iterable[Tuple[str, float]]) -> str:
    lines = [f"{i+1}. {h[0]:<15s}  {h[1]:.2f}" for i, h in enumerate(highscores)]
    return "\n".join(lines)


def format_player_highscores(
    highscores: List[hs.HighscoreStruct], difficulty: Optional[str] = None
) -> List[str]:
    lines = []
    if highscores:
        last_played = format_timestamp(max(h.timestamp for h in highscores))
    else:
        last_played = "<n/a>"
    lines.append(f"Last game played on {last_played}")

    if not difficulty:
        for diff in ["beginner", "intermediate", "expert", "master"]:
            hscores = [h.elapsed for h in highscores if h.difficulty.lower() == diff[0]]
            if hscores:
                best = f"{min(hscores):.2f}"
            else:
                best = "None"
            line = "{}: {}".format(diff.capitalize(), best)
            lines.append(line)
    else:
        lines.append(f"Top {difficulty} times:")
        for h in highscores[:5]:
            line = "{:.2f} ({:.2f} 3bv/s) - {}".format(
                h.elapsed, h.bbbvps, format_timestamp(h.timestamp)
            )
            lines.append(line)

    return lines


def format_kwargs(kwargs: Mapping) -> str:
    return ", ".join(f"{k}={v}" for k, v in kwargs.items())


def format_filters(
    difficulty: Optional[str],
    drag_select: Optional[bool],
    per_cell: Optional[int],
    *,
    no_difficulty=False,
) -> str:
    opts = dict()
    if not no_difficulty:
        if not difficulty:
            difficulty = "combined"
        opts["difficulty"] = difficulty
    if drag_select is not None:
        opts["drag-select"] = "on" if drag_select else "off"
    if per_cell is not None:
        opts["per-cell"] = per_cell
    return format_kwargs(opts)


def format_matchups(matchups: Iterable[Matchup]) -> List[str]:
    return [
        "{} ({:.2f}) vs {} ({:.2f}) - {:.2f}% difference".format(*m) for m in matchups
    ]


def format_timestamp(timestamp: float) -> str:
    return time.strftime("%Y-%m-%d %H:%M", time.localtime(timestamp))



"""
minefield.py - Minefield widgets

April 2018, Lewis Gaul

Exports:
MinefieldWidget
    A minefield widget class, to be packed in a parent container. Receives
    clicks on the cells and makes calls to the backend.
"""

__all__ = ("MinefieldWidget",)

import logging
import sys
from typing import Dict, Optional, Set

from PyQt5.QtCore import QSize, Qt, pyqtSignal
from PyQt5.QtGui import QImage, QMouseEvent, QPainter, QPixmap
from PyQt5.QtWidgets import (
    QApplication,
    QGraphicsScene,
    QGraphicsView,
    QSizePolicy,
    QWidget,
)

from ..core import api
from ..types import (
    CellFlag,
    CellHitMine,
    CellImageType,
    CellMine,
    CellMineType,
    CellNum,
    CellUnclicked,
    CellWrongFlag,
)
from ..typing import Coord_T
from . import state
from .utils import IMG_DIR


logger = logging.getLogger(__name__)


# @@@LG :'(         ... please don't look at the contents of these two functions.
def init_or_update_cell_images(cell_images, size, styles, required=CellImageType.ALL):
    """
    Initialise or update the pixmap images for the minefield cells.

    Arguments:
    cell_images (dict)
        The dictionary to fill with the images.
    size (int)
        The size in pixels to make the image (square).
    required (CellImageType)
        Which images types require updating.
    """
    # @@@LG Currently only allows setting button styles.
    btn_style = styles[CellImageType.BUTTONS]
    if required & CellImageType.BUTTONS:
        cell_images["btn_up"] = make_pixmap("buttons", btn_style, "btn_up.png", size)
        cell_images["btn_down"] = make_pixmap(
            "buttons", btn_style, "btn_down.png", size
        )
        cell_images[CellUnclicked()] = cell_images["btn_up"]
        cell_images[CellNum(0)] = cell_images["btn_down"]

    if required & (CellImageType.BUTTONS | CellImageType.NUMBERS):
        for i in range(1, 19):
            cell_images[CellNum(i)] = make_pixmap(
                "numbers", btn_style, "btn_down.png", size, "num%d.png" % i, 7 / 8
            )

    if required & (CellImageType.BUTTONS | CellImageType.MARKERS):
        for i in range(1, 4):
            cell_images[CellFlag(i)] = make_pixmap(
                "markers", btn_style, "btn_up.png", size, "flag%d.png" % i, 5 / 8
            )
            cell_images[CellWrongFlag(i)] = make_pixmap(
                "markers", btn_style, "btn_up.png", size, "cross%d.png" % i, 5 / 8
            )
            cell_images[CellMine(i)] = make_pixmap(
                "markers", btn_style, "btn_down.png", size, "mine%d.png" % i, 7 / 8
            )
            cell_images[CellHitMine(i)] = make_pixmap(
                "markers", btn_style, "btn_down_hit.png", size, "mine%d.png" % i, 7 / 8
            )


def make_pixmap(img_subdir, style, bg_fname, size, fg_fname=None, propn=1):
    def get_path(subdir, fname, style) -> str:
        base_path = IMG_DIR / subdir
        full_path = base_path / style / fname
        if not full_path.exists():
            logger.warning(f"Missing image file at {full_path}, using standard style")
            full_path = base_path / "standard" / fname
        return str(full_path)

    bg_path = get_path("buttons", bg_fname, style)
    if fg_fname:
        image = QImage(bg_path).scaled(
            size, size, transformMode=Qt.SmoothTransformation
        )
        fg_size = propn * size
        fg_path = get_path(img_subdir, fg_fname, "Standard")
        overlay = QPixmap(fg_path).scaled(
            fg_size, fg_size, transformMode=Qt.SmoothTransformation
        )
        painter = QPainter(image)
        margin = size * (1 - propn) / 2
        painter.drawPixmap(margin, margin, overlay)
        painter.end()
        image = QPixmap.fromImage(image)
    else:
        image = QPixmap(bg_path).scaled(
            size, size, transformMode=Qt.SmoothTransformation
        )
    return image


class MinefieldWidget(QGraphicsView):
    """
    The minefield widget.
    """

    at_risk_signal = pyqtSignal()
    no_risk_signal = pyqtSignal()
    size_changed = pyqtSignal()

    def __init__(
        self,
        parent: Optional[QWidget],
        ctrlr: api.AbstractController,
        state_: state.State,
    ):
        super().__init__(parent)
        logger.info("Initialising minefield widget")
        self._ctrlr: api.AbstractController = ctrlr
        self._state: state.State = state_
        self._cell_images: Dict = {}
        init_or_update_cell_images(self._cell_images, self.btn_size, self._state.styles)

        self.scene = QGraphicsScene()
        self.setScene(self.scene)
        self.setStyleSheet("border: 0px")
        self.setSizePolicy(QSizePolicy.Maximum, QSizePolicy.Maximum)
        self.setMaximumSize(self.sizeHint())

        # Keep track of mouse button states.
        self.mouse_coord = None
        self.both_mouse_buttons_pressed = False
        self.await_release_all_buttons = False
        self.was_double_left_click = False
        self.unflag_on_right_drag = False
        # Set of coords for cells which are sunken.
        self.sunken_cells: Set = set()

        self._ignore_clicks = False
        self.reset()

    @property
    def _board(self):
        # TODO: I don't like this.
        return self._ctrlr.board

    @property
    def x_size(self) -> int:
        return self._state.x_size

    @property
    def y_size(self) -> int:
        return self._state.y_size

    @property
    def btn_size(self) -> int:
        return self._state.btn_size

    # --------------------------------------------------------------------------
    # Qt method overrides
    # --------------------------------------------------------------------------
    def sizeHint(self) -> QSize:
        return QSize(self.x_size * self.btn_size, self.y_size * self.btn_size)

    def mousePressEvent(self, event: QMouseEvent):
        """Handle mouse press events."""

        # Ignore any clicks which aren't the left or right mouse buttons.
        if event.button() not in [Qt.LeftButton, Qt.RightButton]:
            return
        if event.button() == event.buttons():
            self.await_release_all_buttons = False
            self.both_mouse_buttons_pressed = False
        elif self.await_release_all_buttons:
            return
        if self._ignore_clicks or self.was_double_left_click:
            return

        self.mouse_coord = coord = self._coord_from_event(event)

        ## Bothclick
        if event.buttons() & Qt.LeftButton and event.buttons() & Qt.RightButton:
            logger.debug("Both mouse buttons down on cell %s", coord)
            self.both_mouse_buttons_pressed = True
            self.both_buttons_down(coord)
        ## Leftclick
        elif event.button() == Qt.LeftButton:
            logger.debug("Left mouse button down on cell %s", coord)
            self.was_double_left_click = False
            self.left_button_down(coord)
        ## Rightclick
        elif event.button() == Qt.RightButton:
            logger.debug("Right mouse button down on cell %s", coord)
            self.right_button_down(coord)

    def mouseDoubleClickEvent(self, event: QMouseEvent):
        """Handle double clicks."""

        coord = self._coord_from_event(event)

        # Redirect double right-clicks to be two normal clicks.
        if event.button() == Qt.RightButton:
            return self.mousePressEvent(event)
        elif event.button() == Qt.LeftButton and not self.both_mouse_buttons_pressed:
            self.was_double_left_click = True
            self.left_button_double_down(coord)

    def mouseMoveEvent(self, event: QMouseEvent):
        """Handle mouse move events."""

        if self._ignore_clicks:
            return

        coord = self._coord_from_event(event)

        # Return if not the left or right mouse buttons, or if the mouse wasn't
        #  moved to a different cell.
        if (
            not event.buttons() & (Qt.LeftButton | Qt.RightButton)
            or self.await_release_all_buttons
            or coord == self.mouse_coord
        ):
            return

        self.mouse_coord = coord

        ## Double leftclick
        if self.was_double_left_click:
            if event.buttons() == Qt.LeftButton:
                self.left_button_double_move(coord)
            return

        ## Bothclick
        if event.buttons() & Qt.LeftButton and event.buttons() & Qt.RightButton:
            self.both_buttons_move(coord)
        elif not self.both_mouse_buttons_pressed or self._state.drag_select:
            ## Leftclick
            if event.buttons() & Qt.LeftButton:
                self.left_button_move(coord)
            ## Rightclick
            if event.buttons() & Qt.RightButton:
                self.right_button_move(coord)

    def mouseReleaseEvent(self, event: QMouseEvent):
        """Handle mouse release events."""

        if self.await_release_all_buttons and not event.buttons():
            self.await_release_all_buttons = False
            return
        # Ignore any clicks which aren't the left or right mouse buttons.
        if event.button() not in [Qt.LeftButton, Qt.RightButton] or self._ignore_clicks:
            return

        coord = self._coord_from_event(event)

        ## Bothclick (one of the buttons still down)
        if event.buttons() & (Qt.LeftButton | Qt.RightButton):
            logger.debug("Mouse button release on cell %s after both down", coord)
            self.first_of_both_buttons_release(coord)

            if self._state.drag_select and event.button() == Qt.LeftButton:
                # Only right button down - no risk.
                self.no_risk_signal.emit()

        elif not self.both_mouse_buttons_pressed:
            ## Leftclick
            if event.button() == Qt.LeftButton and not self.was_double_left_click:
                logger.debug("Left mouse button release on cell %s", coord)
                self.left_button_release(coord)

        # Reset variables if neither of the mouse buttons are down.
        if not (event.buttons() & (Qt.LeftButton | Qt.RightButton)):
            logger.debug("No mouse buttons down, reset variables")
            self.all_buttons_release()

    # --------------------------------------------------------------------------
    # Mouse click handlers
    # --------------------------------------------------------------------------
    def left_button_down(self, coord: Coord_T) -> None:
        """
        Left mouse button was pressed (single click). Change display and call
        callback functions as appropriate.
        """
        if self._state.drag_select:
            self.at_risk_signal.emit()
            self._ctrlr.select_cell(coord)
        else:
            self._sink_unclicked_cell(coord)

    def left_button_double_down(self, coord: Coord_T) -> None:
        """
        Left button was double clicked. Call callback to remove any flags that
        were on the cell.
        """
        self._ctrlr.remove_cell_flags(coord)

    def left_button_move(self, coord: Coord_T) -> None:
        """
        Left mouse button was moved after a single click. Change display as
        appropriate.
        """
        self._raise_all_sunken_cells()
        self.no_risk_signal.emit()
        if coord is not None:
            self.left_button_down(coord)

    def left_button_double_move(self, coord: Coord_T) -> None:
        """
        Left mouse button moved after a double click.
        """
        if self._state.drag_select:
            self.left_button_double_down(coord)

    def left_button_release(self, coord: Coord_T) -> None:
        """
        Left mouse button was released. Change display and call callback
        functions as appropriate.
        """
        self._raise_all_sunken_cells()
        self.no_risk_signal.emit()
        if not self._state.drag_select and coord is not None:
            self._ctrlr.select_cell(coord)

    def right_button_down(self, coord: Coord_T) -> None:
        """
        Right mouse button was pressed. Change display and call callback
        functions as appropriate.
        """
        self._ctrlr.flag_cell(coord)
        if self._board[coord] == CellUnclicked():
            self.unflag_on_right_drag = True
        else:
            self.unflag_on_right_drag = False

    def right_button_move(self, coord: Coord_T) -> None:
        """
        Right mouse button was moved. Change display as appropriate.
        """
        if coord is not None and self._state.drag_select:
            if self.unflag_on_right_drag:
                self._ctrlr.remove_cell_flags(coord)
            else:
                self._ctrlr.flag_cell(coord, flag_only=True)

    def both_buttons_down(self, coord: Coord_T) -> None:
        """
        Both left and right mouse buttons were pressed. Change display and call
        callback functions as appropriate.
        """
        if not isinstance(self._board[coord], CellMineType):
            for c in self._board.get_nbrs(coord, include_origin=True):
                self._sink_unclicked_cell(c)
        if self._state.drag_select:
            self.at_risk_signal.emit()
            self._ctrlr.chord_on_cell(coord)

    def both_buttons_move(self, coord: Coord_T) -> None:
        """
        Both left and right mouse buttons were moved. Change display as
        appropriate.
        """
        self._raise_all_sunken_cells()
        if not self._state.drag_select:
            self.no_risk_signal.emit()
        if coord is not None:
            self.both_buttons_down(coord)

    def first_of_both_buttons_release(self, coord: Coord_T) -> None:
        """
        One of the mouse buttons was released after both were pressed. Change
        display and call callback functions as appropriate.
        """
        self._raise_all_sunken_cells()
        if not self._state.drag_select:
            self.no_risk_signal.emit()
            if coord is not None:
                self._ctrlr.chord_on_cell(coord)

    def all_buttons_release(self) -> None:
        """
        The second of the mouse buttons was released after both were pressed.
        Change display and call callback functions as appropriate.
        """
        if self._state.drag_select:
            self.no_risk_signal.emit()
        self.mouse_coord = None
        self.both_mouse_buttons_pressed = False
        self.await_release_all_buttons = False
        self.was_double_left_click = False

    # --------------------------------------------------------------------------
    # Other methods
    # --------------------------------------------------------------------------
    def _is_coord_in_grid(self, coord: Coord_T) -> bool:
        return self._board.is_coord_in_grid(coord)

    def _coord_from_event(self, event: QMouseEvent) -> Optional[Coord_T]:
        pos = self.mapToScene(event.pos())

        coord = int(pos.x()) // self.btn_size, int(pos.y()) // self.btn_size
        if not self._is_coord_in_grid(coord):
            return None
        return coord

    def _sink_unclicked_cell(self, coord: Coord_T) -> None:
        """
        Set an unclicked cell to appear sunken.
        """
        if self._state.game_status.finished():
            return
        if self._board[coord] == CellUnclicked():
            self.set_cell_image(coord, "btn_down")
            self.sunken_cells.add(coord)
        if self.sunken_cells:
            self.at_risk_signal.emit()

    def _raise_all_sunken_cells(self) -> None:
        """
        Reset all sunken cells to appear raised.
        """
        while self.sunken_cells:
            coord = self.sunken_cells.pop()
            if self._board[coord] == CellUnclicked():
                self.set_cell_image(coord, "btn_up")

    def reset(self) -> None:
        """Reset all cell images and other state for a new game."""
        logger.info("Resetting minefield widget")
        self.mouse_coord = None
        self.both_mouse_buttons_pressed = False
        self.await_release_all_buttons = True
        for c in self._board.all_coords:
            self.set_cell_image(c, CellUnclicked())

    def set_cell_image(self, coord: Coord_T, state) -> None:
        """
        Set the image of a cell.

        Arguments:
        coord ((x, y) tuple in grid range)
            The coordinate of the cell.
        state
            The cell_images key for the image to be set.
        """
        if state not in self._cell_images:
            logger.error("Missing cell image for state: %s", state)
            return
        x, y = coord
        b = self.scene.addPixmap(self._cell_images[state])
        b.setPos(x * self.btn_size, y * self.btn_size)

    def ignore_clicks(self) -> None:
        self._ignore_clicks = True

    def accept_clicks(self) -> None:
        self._ignore_clicks = False

    def reshape(self, x_size: int, y_size: int) -> None:
        logger.info("Resizing minefield to %sx%s", x_size, y_size)
        self._update_size()
        for c in [(i, j) for i in range(self.x_size) for j in range(self.y_size)]:
            self.set_cell_image(c, CellUnclicked())

    def update_style(self, img_type: CellImageType, style: str) -> None:
        """Update the cell images."""
        logger.info("Updating %s style to '%s'", img_type.name, style)
        init_or_update_cell_images(
            self._cell_images, self.btn_size, self._state.styles, img_type
        )
        for coord in self._board.all_coords:
            self.set_cell_image(coord, self._board[coord])

    def update_btn_size(self, size: int) -> None:
        """Update the size of the cells."""
        assert size == self._state.btn_size
        init_or_update_cell_images(self._cell_images, self.btn_size, self._state.styles)
        for coord in self._board.all_coords:
            self.set_cell_image(coord, self._board[coord])
        self._update_size()

    def _update_size(self) -> None:
        self.setMaximumSize(self.sizeHint())
        self.setSceneRect(
            0, 0, self.x_size * self.btn_size, self.y_size * self.btn_size
        )
        self.size_changed.emit()


if __name__ == "__main__":
    from minegauler.core import GameController
    from minegauler.shared.utils import GameOptsStruct

    _app = QApplication(sys.argv)
    x, y = 6, 4
    ctrlr = GameController(GameOptsStruct(x_size=x, y_size=y))
    state_ = state.State()
    state_.x_size = x
    state_.y_size = y
    state_.btn_size = 100
    mf_widget = MinefieldWidget(None, ctrlr, state_=state_)
    mf_widget.show()
    sys.exit(_app.exec_())
